pub struct TaggedError { /* private fields */ }Expand description
An EncodingError which also displays all the error stack.
This is useful for debugging, because the entire structure tree is displayed.
Implementations
Sourceimpl TaggedError
impl TaggedError
Sourcepub const fn from_stack(err: EncodingError, stack: Stack) -> Self
pub const fn from_stack(err: EncodingError, stack: Stack) -> Self
Constructs a new TaggedError from an encoding error and an
error metadata stack.
If the debug feature is enabled, you can obtain an instance of [Stack]
from an [Encoder] right after an error occurred (encoder.stack).