pub fn decode_bytes_with<'a, R: AsRef<[u8]>, V: Decode<Slice<'a>>>(
bytes: &'a R,
context: Context<'_>,
) -> EncodingResult<V>Expand description
Decodes the given value by constructing an encoder on the fly and using it to wrap a byte slice.
Uses the given Context